#!/usr/bin/env python3
# coding: utf-8

# Autorship information
__author__ = "Samuel Murail, Maxence Delaunay, Damien Espana"
__copyright__ = "Copyright 2020, RPBS"
__credits__ = ["Damien Espana", "Maxence Delaunay", "Samuel Murail"]
__license__ = "GNU General Public License v2.0"
__version__ = "2.0.0"
__maintainer__ = "Samuel Murail"
__email__ = "samuel.murail@u-paris.fr"
__status__ = "Prototype"
